"""OpenToken utility functions
"""

import json
import re
from collections import OrderedDict


def validate_cipher_suite_id(cipher_suite_id):
    """Validates that a CipherSuite conforms to the proper format.

    Args:
        cipher_suite_id (int): CipherSuite id.

    Returns:
        (int): The original CipherSuite id.

    """
    if not isinstance(cipher_suite_id, int):
        raise TypeError("CipherSuite ID must be of type int.")
    if not 0 <= cipher_suite_id <= 3:
        raise ValueError("Invalid CipherSuite.")
    return cipher_suite_id


def validate_password(password):
    """Validate and reformat password argument.

    Args:
        password (str or bytes): Encryption password.

    Returns:
        str or bytes: The original password string.

    """
    if password is None:
        password = ""

    if not isinstance(password, str) and not isinstance(password, bytes):
        raise TypeError("Invalid password type")

    return password


def validate_payload(payload):
    """Validate that the payload is of type OrderedDict.
    If the payload is of type str, then it assumes that the string is
    able to be parsed via json.

    Args:
        payload (str or OrderedDict): Payload object

    Returns:
        OrderedDict: Original payload object as an OrderedDict.

    """
    if isinstance(payload, str):
        payload = json.JSONDecoder(
            object_pairs_hook=OrderedDict
        ).decode(payload)
    if not isinstance(payload, OrderedDict):
        raise TypeError("Payload must be of type OrderedDict.")
    return payload


def reformat_to_otk_b64(token):
    """Reformat a base64 encoded token to OpenToken standards,
    which replaces all base64 padding characters "=" with "*".

    Args:
        token (str): The base64 encoded token to reformat.

    Returns:
        (str): The reformatted base64 encoded string.

    """
    if re.match(r"^.*={2}$", token):
        token = token[:-2] + "**"
    elif re.match(r"^.*=$", token):
        token = token[:-1] + "*"
    return token


def reformat_from_otk_b64(token):
    """Reformat a base64 encoded token from OpenToken standards,
    which replaces all base64 padding characters "*" with "=".

    Args:
        token (str): The base64 encoded token to reformat.

    Returns:
        (str): The reformatted base64 encoded string.

    """
    if re.match(r"^.*\*{2}$", token):
        token = token[:-2] + "=="
    elif re.match(r"^.*\*$", token):
        token = token[:-1] + "="
    return token


def ordered_dict_to_otk_str(otk_dict):
    """Converts an OrderedDict to a OpenToken string.

    Args:
        otk_dict (OrderedDict): OrderedDict representation of the token.

    Returns:
        str: String representation of the token.

    """
    otk_list = []
    for k, v in otk_dict.items():
        otk_list.append("{0}={1}".format(k, v))
    return "\n".join(otk_list)


def otk_str_to_ordered_dict(otk_str):
    """Converts an OrderedDict to a OpenToken string.

    Args:
        otk_str (str): String representation of the token.

    Returns:
        OrderedDict: OrderedDict representation of the token.

    """
    items = otk_str.split("\n")
    pairs = [tuple(line.split("=")) for line in items]
    return OrderedDict(pairs)



"""Token helper methods
"""

import base64
from collections import OrderedDict
from zlib import compress, decompress

from Cryptodome.Cipher import AES, DES3
from Cryptodome.Hash import SHA1, HMAC
from Cryptodome.Random import get_random_bytes
from Cryptodome.Util.Padding import pad, unpad

from . import _ciphersuite, _utils


def encode(payload, cipher_suite_id, password=None):
    """Generate an OpenToken from a given payload.

    OTK uses a simple, line-based format for encoding the key-value pairs
    in the payload. The format is encoded with UTF-8 and thus is
    guaranteed to support the transport of multi-byte characters.

    Args:
        payload (OrderedDict): Data to encrypt.
        cipher_suite_id (int): Cipher suite id.
        password (str): Password used for encryption/decryption.

    """
    payload = _utils.validate_payload(payload)
    cipher_suite_id = _utils.validate_cipher_suite_id(cipher_suite_id)
    password = _utils.validate_password(password)

    cipher = _ciphersuite.CIPHERS[cipher_suite_id]

    otk_version = 1
    encryption_key = _ciphersuite.generate_key(password, cipher_suite_id)
    iv_length = cipher["iv_length"]
    payload = bytes(_utils.ordered_dict_to_otk_str(payload), "utf-8")
    iv = get_random_bytes(iv_length)

    if cipher_suite_id == 0:
        hmac = SHA1.new()
    else:
        hmac = HMAC.new(encryption_key, digestmod=SHA1)
    hmac.update(bytearray([otk_version]))
    hmac.update(bytearray([cipher_suite_id]))
    if iv_length > 0:
        hmac.update(iv)
    hmac.update(payload)
    hmac_digest = hmac.digest()

    zipped_data = compress(payload)

    if cipher_suite_id == 3:
        cipher_type = DES3
    else:
        cipher_type = AES
    cipher = cipher_type.new(encryption_key, cipher_type.MODE_CBC, iv=iv)
    payload_cipher_text = cipher.encrypt(
        pad(zipped_data, cipher_type.block_size)
    )

    otk_buffer = bytearray("OTK", "utf-8")  #: OTK literal
    otk_buffer.append(1)  #: Version identifier
    otk_buffer.append(cipher_suite_id)  #: Cipher suite identifier
    otk_buffer.extend(hmac_digest)  #: SHA-1 HMAC
    otk_buffer.append(iv_length)  #: IV Length
    if iv_length > 0:
        otk_buffer.extend(iv)  #: IV
    otk_buffer.append(0)  #: Key info length
    otk_buffer.extend(
        int(len(payload_cipher_text)).to_bytes(
            2, byteorder="big", signed=False
        )
    )  #: Payload length
    otk_buffer.extend(payload_cipher_text)  #: Payload

    otk = base64.urlsafe_b64encode(otk_buffer).decode("utf-8")
    return _utils.reformat_to_otk_b64(otk)


def decode(otk, cipher_suite_id, password=None):
    """Decode an OpenToken.

    Args:
        otk (str): Base64 encoded OpenToken with "*" padding chars.
        cipher_suite_id (int): Cipher suite id.
        password (str): Password used for encryption/decryption.

    """
    cipher_suite_id = _utils.validate_cipher_suite_id(cipher_suite_id)
    password = _utils.validate_password(password)

    decryption_key = _ciphersuite.generate_key(password, cipher_suite_id)
    otk = _utils.reformat_from_otk_b64(otk)
    read_index = 0
    otk = bytearray(base64.urlsafe_b64decode(otk))

    #: Validate the OTK header literal
    otk_header = otk[read_index:read_index + 3].decode()
    read_index += 3
    if otk_header != "OTK":
        raise ValueError(
            "Invalid token header literal: {0}".format(otk_header)
        )

    #: Validate version
    otk_version = int.from_bytes(otk[read_index:read_index + 1], "big")
    read_index += 1
    if otk_version != 1:
        raise ValueError("Invalid OTK version.")

    #: Validate CipherSuite id
    otk_cipher_suite_id = int.from_bytes(otk[read_index:read_index + 1], "big")
    read_index += 1
    if otk_cipher_suite_id != cipher_suite_id:
        raise ValueError(
            "CipherID, {0}, doesn't match the encoding cipher, {1}.".format(
                otk_cipher_suite_id, cipher_suite_id
            )
        )

    #: Extract cipher, mac and iv information
    hmac = otk[read_index:read_index + 20]
    read_index += 20
    iv_length = int.from_bytes(otk[read_index:read_index + 1], "big")
    read_index += 1
    iv = None
    if iv_length > 0:
        iv = otk[read_index:read_index + iv_length]
        read_index += iv_length

    #: Extract the Key Info (if present) and select a key for decryption
    key_info_length = int.from_bytes(otk[read_index:read_index + 1], "big")
    read_index += 1
    key_info = None
    if key_info_length > 0:
        key_info = otk[read_index:read_index + key_info_length]
        read_index += key_info_length

    #: Decrypt the payload cipher-text using the selected cipher suite
    payload_cipher_text = None
    payload_length = int.from_bytes(
        otk[read_index:read_index + 2], "big", signed=False
    )
    read_index += 2
    payload_cipher_text = otk[read_index:read_index + payload_length]

    if cipher_suite_id == 3:
        cipher_type = DES3
    else:
        cipher_type = AES
    cipher = cipher_type.new(decryption_key, cipher_type.MODE_CBC, iv=iv)
    try:
        zipped_data = unpad(
            cipher.decrypt(payload_cipher_text), cipher_type.block_size
        )
    except ValueError:
        raise ValueError("Error decrypting token.")

    #: Decompress the decrypted payload in accordance with RFC1950 and RFC1951
    payload = decompress(zipped_data)

    #: Initialize an HMAC using the SHA-1 algorithm and the following data -
    #: OTK Version, Cipher Suite Value, IV value, Key info value (if present)
    if cipher_suite_id == 0:
        hmac_test = SHA1.new()
    else:
        hmac_test = HMAC.new(decryption_key, digestmod=SHA1)
    hmac_test.update(bytearray([otk_version]))
    hmac_test.update(bytearray([cipher_suite_id]))
    if iv_length > 0:
        hmac_test.update(iv)
    if key_info:
        hmac_test.update(key_info)
    hmac_test.update(payload)

    #: Compare reconstructed HMAC with original HMAC
    if hmac_test.hexdigest() != hmac.hex():
        raise ValueError("HMAC does not match.")

    return _utils.otk_str_to_ordered_dict(payload.decode())



"""OpenToken module for Python
"""

import datetime
from collections import OrderedDict

import dateutil.parser

from . import _token


class OpenToken:
    """API class for generating and reading OpenTokens.

    Args:
        password (str): Password used for encryption/decryption.
        cipher_suite_id (int): Cipher suite id.
        token_tolerance (int): Token tolerance.
        token_lifetime (int): Token lifetime.
        token_renewal (int): Token renewal.

    """

    def __init__(self, password=None, cipher_suite_id=2, token_tolerance=120,
                 token_lifetime=300, token_renewal=43200):
        self.cipher_suite_id = cipher_suite_id
        self.password = password
        self.token_tolerance = token_tolerance
        self.token_lifetime = token_lifetime
        self.token_renewal = token_renewal

    def parse_token(self, otk_str):
        """Parse an OpenToken and apply basic validation checks.

        Args:
            otk_str (str): The raw base64 encoded token string.

        Returns:
            OrderedDict: The key-value token pairs.

        """
        parsed_token = _token.decode(
            otk_str, self.cipher_suite_id, self.password
        )

        if "subject" not in parsed_token.keys():
            raise ValueError("OpenToken missing 'subject'.")

        not_before = dateutil.parser.isoparse(parsed_token['not-before'])
        not_on_or_after = dateutil.parser.isoparse(
            parsed_token['not-on-or-after']
        )
        renew_until = dateutil.parser.isoparse(parsed_token['renew-until'])
        now = datetime.datetime.now(datetime.timezone.utc)
        tolerance = now + datetime.timedelta(seconds=self.token_tolerance)

        if not_before > not_on_or_after:
            raise ValueError(
                "Logical error in 'not-before' and 'not-on-or-after'."
            )

        if not_before > now and not_before > tolerance:
            raise ValueError("Must not use this token before {0}.".format(
                parsed_token['not-before']
            ))

        if now > not_on_or_after:
            raise ValueError("This token has expired as of {0}.".format(
                parsed_token['not-on-or-after']
            ))

        if now > renew_until:
            raise ValueError(
                "This token is past its renewal limit, {0}.".format(
                    parsed_token['renew-until']
                )
            )

        return parsed_token

    def create_token(self, otk_pairs):
        """Create an OpenToken from an object of key-value pairs to encode.

        Args:
            otk_pairs (list): The key-value token pairs as a list of tuples.

        Returns:
            str: The raw base64 encoded token string.

        """
        otk_dict = OrderedDict(otk_pairs)

        if "subject" not in otk_dict.keys():
            raise ValueError("OpenToken missing 'subject'.")

        now = datetime.datetime.now(datetime.timezone.utc)
        expiry = now + datetime.timedelta(seconds=self.token_lifetime)
        renew_until = now + datetime.timedelta(seconds=self.token_renewal)

        otk_dict['not-before'] = now.isoformat()
        otk_dict['not-on-or-after'] = expiry.isoformat()
        otk_dict['renew-until'] = renew_until.isoformat()

        return _token.encode(otk_dict, self.cipher_suite_id, self.password)



"""CipherSuite helper module
"""

from Cryptodome.Protocol.KDF import PBKDF2

from . import _utils

CIPHERS = [
    {
        "id": 0,
        "name": None,
        "cipher": None,
        "key_size": 0,
        "mode": None,
        "padding": None,
        "iv_length": 0
    },
    {
        "id": 1,
        "name": "aes-256-cbc",
        "cipher": "AES",
        "key_size": 256,
        "mode": "CBC",
        "padding": "PKCS 5",
        "iv_length": 16
    },
    {
        "id": 2,
        "name": "aes-128-cbc",
        "cipher": "AES",
        "key_size": 128,
        "mode": "CBC",
        "padding": "PKCS 5",
        "iv_length": 16
    },
    {
        "id": 3,
        "name": "3des",
        "cipher": "3DES",
        "key_size": 168,
        "mode": "CBC",
        "padding": "PKCS 5",
        "iv_length": 8
    }
]


def generate_key(password, cipher_suite_id, salt=None):
    password = _utils.validate_password(password)
    cipher_suite_id = _utils.validate_cipher_suite_id(cipher_suite_id)

    if cipher_suite_id == 0:
        return None

    salt = salt or bytearray([0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0])
    cipher_suite = CIPHERS[cipher_suite_id]

    return PBKDF2(
        password,
        salt,
        dkLen=cipher_suite["key_size"] // 8,
    )



import setuptools

with open("README.md", "r") as f:
    readme = f.read()

packages = ['opentoken']

with open("requirements.txt", "r") as f:
    requires = f.read().split("\n")

test_requirements = [
    "pytest==5.3.2"
]

setuptools.setup(
    name="opentoken",
    version="2.1.0",
    description="OpenToken support for python.",
    long_description=readme,
    long_description_content_type='text/markdown',
    author="Jason Yoon",
    url="https://github.com/yoonjesung/opentoken-python",
    packages=packages,
    package_dir={"opentoken": "opentoken"},
    python_requires=">=3.5",
    install_requires=requires,
    license="MIT",
    tests_require=test_requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
